#include <stdio.h>
#include <string.h>
#include <simcli.h>

#define esousr  "t023es"
#define esopwd  "onedatabase"

#define PHIMIN   0.0
#define PHIMAX  10.0
#define MAGMIN   0.0
#define MAGMAX  25.0
#define MAXOUT  1000

char *http, *getenv();

main(argc, argv) int argc; char **argv;
{

	int hh ;
	int l, nitems, nout, fmt;
	char *p;
	char line[BUFSIZ], name[BUFSIZ], question[BUFSIZ], option[BUFSIZ];
	double ra, dec, r1, r2, m1, m2;

	/* check here if we are called from an HTTP deamon */
	if( (http=getenv("QUERY_STRING")) || (http=getenv("PATH_INFO")) ){
	     printf("Content-Type: text/plain\n\n");
	     fflush(stdout);
	}

	if(!http) {
	     printf("Connecting to SIMBAD ... ") ; fflush(stdout);
	}
	hh = simbad_connect("", "", esousr, esopwd) ;
	if (hh < 0)
	{
	     char *text = simbad_error(hh) ;
	     printf("*** Failed to connect: %s\n",text) ;
	     exit(1) ;
	}
	if(!http) {
	     printf("connected \n"); fflush(stdout);
	}

	if(http){
	     strncpy(line, http, sizeof(line)-1);
	     l = 1;
	} else {
	     l = read(0,line,sizeof(line));
	     line[--l] = '\0'; /* strip out the newline */
	}
	while(l > 0){

	     m1 = MAGMIN;
	     m2 = MAGMAX;
	     nout = MAXOUT; 
	     name[0] = question[0] = option[0] = '\0';
	     r1 = r2 = 0.;
	     fmt=8;

	     if(!parse_cone(line,name,&ra,&dec,&r1,&r2,&m1,&m2,&nout,&fmt)){
		  printf(
		  "*** You must supply a coordinate center or object name.\n");
		  fflush(stdout);
		  goto next;
	     }

	     if(r2 > 0.) {
		  char sradius[128];

		  dtos(r2, sradius, 2);
		  sprintf(option, "cooradius=%s", sradius);
	     }

	     if(*name)
		  sprintf(question, "%s%s", (r2>0.)? "~" : "", name);
	     else {
		  char sra[128], sdec[128];

		  dtos(ra, sra, 2);
		  dtos(dec, sdec, 2);
		  sprintf(question, "%s %s", sra+1, sdec);
	     }

             if(!http) {
                printf("Request: %s \n",question); 
                printf("Option: %s \n",option); fflush(stdout);
             }

	     nitems = simbad_query(hh, question, option);
	     if (nitems < 0){
		  printf("*** %s\n",simbad_error(hh)) ;
		  fflush(stdout);
	     } else {
		  if(nitems > nout) nitems = nout;
		  if(fmt == 0){
		       while(nitems--){
			    printf("\nInfo for %s\n", question); 
			    printf("------------------------------------\n");
			    fflush(stdout);
			    while (sim_next_all(hh));
		       }
		  } else if(r2 > 0.) {
		       printf(
		    "Id         \t ra       \tdec      \tmv \tmb\tObj-Type\tMORE\n"); 
		       printf(
		    "-----------\t ---------\t---------\t---\t---\t--------\n");
		       fflush(stdout);
		       while (nitems-- && sim_next_coo(hh));
		  } else {
		       printf("Id         \t ra       \tdec\n"); 
		       printf("-----------\t ---------\t----------\n");
		       fflush(stdout);
		       sim_name(hh);
		  }
		  printf("[EOD]\n"); fflush(stdout);
	     }
	     
	   next:
	     if(http) {
		  l = 0;
	     } else {
		  l = read(0,line,sizeof(line));
		  line[--l] = '\0'; /* strip out the newline */
	     }
	}
	simbad_disconnect(hh) ;
}

int parse_cone(line,name,ra,dec,r1,r2,m1,m2,nout,fmt)
char *line, *name;
double *ra, *dec, *r1, *r2, *m1, *m2;
int *nout, *fmt;
{
	char *pp, *cc, *qq, *s;
	int rh, rm, ddd, dm, n, ii, jj, center;
	double rs, ds, x, y, atof();
	void unescape_url();

	unescape_url(line);

	center = 0;
	rs = ds = 0.;
	rh = rm = ddd = dm = ii = jj = 0;
	pp = strtok(line, ";&/");
	while(pp && *pp){
	     while(*pp && *pp == ' ') pp++;
	     cc = pp;
	     while(*pp && isalpha(*pp)) pp++;
	     switch(*cc){
		case 'f':
		  /* format [0-3] */
		  *fmt = atoi(pp);
		  break;
		case 'n':
		  /* nout  */
		  while(*pp && !isdigit(*pp)) pp++;
		  *nout = atoi(pp);
		  break;
		case 'o':
		  /* object name  */
		  while(*pp && *pp != '=') pp++;
		  if(*pp == '=') strcpy(name, ++pp);
		  center = 1;
		  break;
		case 's':
		  /* sort field [1-12]|{r|d|m|-} */
		  break;
		case 'r':
		  /* radius [1] 2 in arcmin */
		  while(*pp && !isalnum(*pp)) pp++;
		  if((s=strchr(pp, ','))) *s = ' ';
		  x = y = 0.;
		  ii = jj = 0;
		  sscanf(pp, "%lf %lf", &x, &y);
		  sscanf(pp, "%d %d", &ii, &jj);
		  if( ii > 0 && (int)x != ii ) x = (double)ii;
		  if( jj > 0 && (int)y != jj ) y = (double)jj;
		  if( y == 0. ){ 
		       *r2 = (x == 0.)? PHIMAX/60. : x/60.; 
		       *r1 = 0.;
		  } else {
		       *r1 = x/60.;
		       *r2 = y/60.;
		  }
		  break;
		  
		case 'q': exit(0);

		case 'F':
		  /* Flags: filter multiple stars out */
		  break;

		case 'm':
		  if(strstr(cc, "mime")){
		       while(*pp && *pp != '=') pp++;
		       if(strstr(pp, "tsv"))
			 *fmt = 8;
		       else if(strstr(pp, "full-rec"))
			 *fmt = 0;
		  } else {
		       /* limit mag 1 [2] */ 
		       while(*pp && !isalnum(*pp)) pp++;
		       if((s=strchr(pp, ','))) *s = ' ';
		       sscanf(pp, "%f %f", m1, m2);
		       sscanf(pp, "%d %d", &ii, &jj);
		       if( ii > 0 )
			 *m1 = ( (int)*m1 == ii)? *m1 : (float)ii;
		       if( jj > 0 )
			 *m2 = ( (int)*m2 == jj)? *m2 : (float)jj;
		  }
		  break;
		       
		default:
		  if(isdigit(*cc) 
		     || *cc == '-' || *cc == '+' || *cc == 'c'){
		       /* center coordinates: 
			  hh mm ss.s +/-dd mm ss.s or
			  h.hhhh +/-d.dddd
			  */
		       if(to_d(pp, ra, dec)) center = 1;
		       else {
			    printf("*** Unable to parse RA/DEC: %s\n", pp);
			    fflush(stdout);
		       }

		  } else if(*cc > ' ') { 
		       help();
		       if(*cc != '?' && tolower(*cc) != 'h'){
			    printf("*** Unknown command: %s\n", cc);
			    fflush(stdout);
		       }
		       return(0); 
		  }
		  break;
	     }
	     pp = strtok((char *)0, ";/&");
	}

	return(center);
}

int sim_name(hh)
int hh;
{

     int stat, ndata ;
     char *p, *q, buff[BUFSIZ];

     stat = simbad_retrieve(hh,0) ;
     if (stat < 0){
	  printf("*** SIMBAD retrieve error\n") ;
	  fflush(stdout);
	  return(stat);
     } else if(stat == 0) return(stat);

     buff[0] = '\0';
     if(! simbad_findata(hh, "I.0", "")) return(0);
     p = simbad_getdata(hh,1);

     if(! simbad_findata(hh, "J", "")) return(0);
     q = simbad_getdata(hh, 1);
     if(p && q){
	  *(q+12) = '\t'; *(q+24) = '\0';
	  sprintf(buff, "%s\t%s", p+1, q);
     } else return(0);

     printf("%s\n", buff); fflush(stdout);

     return(1);
}

int sim_next_coo(hh)
int hh;
{
     double atof();
     int stat, ndata ;
     char *id, *p, *q, buff[BUFSIZ], more[BUFSIZ], s[BUFSIZ];

     stat = simbad_retrieve(hh,0) ;
     if (stat < 0){
	  printf("*** SIMBAD retrieve error\n") ;
	  fflush(stdout);
	  return(stat);
     } else if(stat == 0) return(stat);

     buff[0] = '\0';
     if(! simbad_findata(hh, "I.0", "")) return(0);
     id = simbad_getdata(hh,1);

     if(! simbad_findata(hh, "J", "")) return(0);
     q = simbad_getdata(hh, 1);
     if(id && q) {
	  *(q+12) = '\t'; *(q+24) = '\0';
	  sprintf(buff, "%s\t%s\t", id+1, q);
     } else return(0);

     if(simbad_findata(hh, "M.V", "")){
	  p = simbad_getdata(hh,1);
	  sprintf(s, "%.2lf", atof(p));
	  strcat(buff, s);
     }
     strcat(buff, "\t");

     if(simbad_findata(hh, "M.B", "")){
	  p = simbad_getdata(hh,1);
	  sprintf(s, "%.2lf", atof(p));
	  strcat(buff, s);
     }
     strcat(buff, "\t");

     if(simbad_findata(hh, "C", "")){
	  p = simbad_getdata(hh,1);
	  strcat(buff, p);
     }
     strcat(buff, "\t");

     for(p=id+1, q=s; *p ; p++)
       if(*p == ' ') {
	    *q++ = '%'; *q++ = '2'; *q++ = '0';
       } else {
	    *q++ = *p;
       }
     *q = '\0';
     if(http) {
	 p = getenv("SERVER_PORT");
	 if(p && (stat=atoi(p)) != 80)
	     sprintf(more, "M=http://%s:%d%s?o=%s/mime-type=full-rec", 
		     getenv("SERVER_NAME"), stat, 
		     getenv("SCRIPT_NAME"), s);
	 else
	     sprintf(more, "M=http://%s%s?o=%s/mime-type=full-rec", 
		     getenv("SERVER_NAME"), getenv("SCRIPT_NAME"), s);
	 strcat(buff, more);
     }

     printf("%s\n", buff); fflush(stdout);

     return(1);
}

int sim_next_all(hh)
int hh;
{

     int stat, ndata, cats;
     char *astrotyplist, *typ, *p ;

     stat = simbad_retrieve(hh,0) ;
     if (stat < 0){
	    printf("*** SIMBAD retrieve error: stat=%d\n", stat) ;
	    fflush(stdout);
	    return(stat);
     } else if(stat == 0) return(stat);

     cats = 0;
     astrotyplist = simbad_telldata(hh);
     for (typ = strtok(astrotyplist," ") ;
	  typ != NULL ;
	  typ = strtok(NULL," "))
       {
	    ndata = simbad_findata(hh,typ,"");
	    if(ndata == 0) continue;
	    switch(*typ){
	       default:
		 break;
	       case 'C':
		 if( (p=simbad_getdata(hh,0)) )
		   printf("%17s: %s\n", "Object Type", p);
		 break;
	       case 'J':
		 if( (p=simbad_getdata(hh,0)) )
		   printf("%17s: %s\n", "J2000 [deg]", p);
		 break;
	       case 'P':
		 if( (p=simbad_getdata(hh,0)) )
		   printf("%17s: %s\n", "P Motions [as/y]", p);
		 break;
	       case 'M':
		 if( (p=simbad_getdata(hh,0)) ){
		      *typ = 'm'; *(typ+2) = tolower( *(typ+2) );
		      printf("%17s: %s\n", typ, p);
		 }
		 break;
	       case 'S':
		 if( (p=simbad_getdata(hh,0)) )
		   printf("%17s: %s\n", "Spectral Type", p);
		 break;
	       case 'D':
		 if( (p=simbad_getdata(hh,0)) )
		   printf("%17s: %s\n", "Dimensions", p);
		 break;
	       case 'T':
		 if( (p=simbad_getdata(hh,0)) )
		   printf("%17s: %s\n", "Morph. Type", p);
		 break;
	       case 'I':
		 if(*(typ+1) == '.') break;
		 printf("%17s:\n", "Id's");
		 while( (p=simbad_getdata(hh,0)) ) 
		   printf("%17s  %s\n", " ", p);
		 break;
	       case 'O':
		 if(!cats) {
		      printf("%17s:\n", "Cat measurements");
		      cats++;
		 }
		 printf("%17s   %d %s\n", " ", ndata, typ+2);
		 break;
	       case 'B':
		 printf("\n %d Bibliographical references available.\n", 
			ndata);
		 break;
	    }
       }
     fflush(stdout);

     return(1);
}

help()
{
     static char helptxt[] = 
"Searching SIMBAD by coordinates or name resolving.\n\
Enter a sequence of commands separated with slashes (/) all in one line.\n\
\n\
command    parameters      format/units\n\
\n\
center     ra +|-dec       free format\n\
radius     r1[,r2]         arcmin (default: 0,10')\n\
mag        mag1[,mag2]     limits in magnitude (mag1<mag2)\n\
nout       no. of lines    Number of lines in output (Max = default = 1000).\n\
\n\
Commands can abbreviated to first letter.\n";

	write(1,helptxt,sizeof(helptxt));
}

void unescape_url(url)
char *url;
{
    register int x,y;
    char x2c();

    for(x=0;url[x];x++)
        if(url[x] == '%')
            url[x+1] = x2c(&url[x+1]);

    for(x=0,y=0;url[y];++x,++y) {
        if((url[x] = url[y]) == '%') {
            url[x] = url[y+1];
            y+=2;
        }
    }
    url[x] = '\0';
}

char x2c(what)
char *what;
{
    register char digit;

    digit = ((what[0] >= 'A') ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
    digit *= 16;
    digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
    return(digit);
}


int to_d(str,ra,dec)
/*+++++++++++
.PURPOSE Parse RA/DEC into decimal degrees.
.REMARKS Input can be hh mm ss.s +/-dd mm ss.s or d.hhhh +/-d.dddd
.RETURNS 1 OK | 0 not parseable
------------*/
     char *str;
     double *ra, *dec;
{
     char *pp, *qq, *s;
     double dd, dm, ds, fabs();
     int i, sign;

     pp = str;
     for(qq=pp;	 *qq && *qq != '+' && *qq != '-' && *qq != ',';  qq++);
     s = qq;
     sign = 1;
     if(*s == '-') {
	  s++;
	  sign = -1;
     }
     for(; *s ; s++)
       if(!isdigit(*s) && *s != '.') *s = ' ';
     
     i = sscanf(qq, "%lf %lf %lf", &dd, &dm, &ds);
     if( i == 0 ) return(0);
     if( i == 1 ) {
	  *dec = dd;
     } else {
	  *dec = fabs(dd) + dm/60. + ds/3600.;
	  *dec *= sign;
     }
     
     *qq = '\0';
     for(s=pp; *s ; s++)
       if(!isdigit(*s) && *s != '.') *s = ' ';
     
     i = sscanf(pp, "%lf %lf %lf", &dd, &dm, &ds);
     if( i == 0 || dd < 0. ) return(0);
     if( i == 1 ) {
	  *ra = dd;
	  *ra = *ra/15.;
     } else {
	  *ra = fabs(dd) + dm/60. + ds/3600.;
     }
}

/*==================================================================
** NAME         :dtos.c
** TYPE         :void
** DESCRIPTION  :transforms a double into a string (base 60)
**              :
** INPUT        :double *deci [,int prec]
**              :
** OUTPUT       :char *string
**              :
** AUTHOR       :apm
** DATE         :06/91; 09/91;
** LIB          :pm.a  - use cco, libad
*=================================================================*/

dtos(deci,string,prec)
	double deci;
	char *string;
	int prec;
{
	char sign;
	int h,m;
	double dd,s,md,fabs();

	sign='+';
	if(deci < 0.0) sign='-';
	dd= fabs(deci)+0.0000000001;
	h=(int)dd;
	md=(dd-h)*60.;
	m=(int)md;
	s=(md-m)*60.;
	if(prec == 3)
	sprintf(string,"%c%02d %02d %06.3f",sign,h,m,s);
	else if(prec == 1)
	sprintf(string,"%c%02d %02d %04.1f",sign,h,m,s);
	else if(prec == 0)
	sprintf(string,"%c%02d %02d %02.0f",sign,h,m,s);
	else if(prec == -1)
	sprintf(string,"%c%02d %04.1f",sign,h,md);
	else
	sprintf(string,"%c%02d %02d %05.2f",sign,h,m,s);
	return;
}




